@@ -1,5 +1,23 @@
 module Curses
   COLOR_DEFAULT = -1
+
+  NUM_COLORS = `tput colors`.to_i
+  MAX_PAIRS = `tput pairs`.to_i
+
+  def self.color! name, value
+    const_set "COLOR_#{name.to_s.upcase}", value
+  end
+
+  ## numeric colors
+  Curses::NUM_COLORS.times { |x| color! x, x }
+
+  if Curses::NUM_COLORS == 256
+    ## xterm 6x6x6 color cube
+    6.times { |x| 6.times { |y| 6.times { |z| color! "c#{x}#{y}#{z}", 16 + z + 6*y + 36*x } } }
+
+    ## xterm 24-shade grayscale
+    24.times { |x| color! "g#{x}", (16+6*6*6) + x }
+  end
 end
 
 module Redwood
@@ -7,12 +25,6 @@ module Redwood
 class Colormap
   @@instance = nil
 
-  CURSES_COLORS = [Curses::COLOR_BLACK, Curses::COLOR_RED, Curses::COLOR_GREEN,
-                   Curses::COLOR_YELLOW, Curses::COLOR_BLUE,
-                   Curses::COLOR_MAGENTA, Curses::COLOR_CYAN,
-                   Curses::COLOR_WHITE, Curses::COLOR_DEFAULT]
-  NUM_COLORS = (CURSES_COLORS.size - 1) * (CURSES_COLORS.size - 1)
-
   DEFAULT_COLORS = {
     :status => { :fg => "white", :bg => "blue", :attrs => ["bold"] },
     :index_old => { :fg => "white", :bg => "default" },
@@ -56,24 +68,35 @@ class Colormap
   def initialize
     raise "only one instance can be created" if @@instance
     @@instance = self
-    @entries = {}
     @color_pairs = {[Curses::COLOR_WHITE, Curses::COLOR_BLACK] => 0}
     @users = []
     @next_id = 0
+    reset
     yield self if block_given?
+  end
+
+  def reset
+    @entries = {}
+    @highlights = { :none => highlight_sym(:none)}
     @entries[highlight_sym(:none)] = highlight_for(Curses::COLOR_WHITE,
                                                    Curses::COLOR_BLACK,
                                                    []) + [nil]
   end
 
-  def add sym, fg, bg, attr=nil, opts={}
+  def add sym, fg, bg, attr=nil, highlight=nil
     raise ArgumentError, "color for #{sym} already defined" if @entries.member? sym
-    raise ArgumentError, "color '#{fg}' unknown" unless CURSES_COLORS.include? fg
-    raise ArgumentError, "color '#{bg}' unknown" unless CURSES_COLORS.include? bg
+    raise ArgumentError, "color '#{fg}' unknown" unless (-1...Curses::NUM_COLORS).include? fg
+    raise ArgumentError, "color '#{bg}' unknown" unless (-1...Curses::NUM_COLORS).include? bg
     attrs = [attr].flatten.compact
 
     @entries[sym] = [fg, bg, attrs, nil]
-    @entries[highlight_sym(sym)] = opts[:highlight] ? @entries[opts[:highlight]] : highlight_for(fg, bg, attrs) + [nil]
+
+    if not highlight
+      highlight = highlight_sym(sym)
+      @entries[highlight] = highlight_for(fg, bg, attrs) + [nil]
+    end
+
+    @highlights[sym] = highlight
   end
 
   def highlight_sym sym
@@ -116,7 +139,7 @@ class Colormap
   end
 
   def color_for sym, highlight=false
-    sym = highlight_sym(sym) if highlight
+    sym = @highlights[sym] if highlight
     return Curses::COLOR_BLACK if sym == :none
     raise ArgumentError, "undefined color #{sym}" unless @entries.member? sym
 
@@ -127,7 +150,7 @@ class Colormap
     if(cp = @color_pairs[[fg, bg]])
       ## nothing
     else ## need to get a new colorpair
-      @next_id = (@next_id + 1) % NUM_COLORS
+      @next_id = (@next_id + 1) % Curses::MAX_PAIRS
       @next_id += 1 if @next_id == 0 # 0 is always white on black
       id = @next_id
       debug "colormap: for color #{sym}, using id #{id} -> #{fg}, #{bg}"
@@ -160,48 +183,35 @@ class Colormap
       Redwood::load_yaml_obj Redwood::COLOR_FN
     end
 
-    error = nil
-    Colormap::DEFAULT_COLORS.each_pair do |k, v|
-      fg = Curses.const_get "COLOR_#{v[:fg].upcase}"
-      bg = Curses.const_get "COLOR_#{v[:bg].upcase}"
-      attrs = v[:attrs] ? v[:attrs].map { |a| Curses.const_get "A_#{a.upcase}" } : []
-
-      if user_colors && (ucolor = user_colors[k])
-        if(ufg = ucolor[:fg])
-          begin
-            fg = Curses.const_get "COLOR_#{ufg.upcase}"
-          rescue NameError
-            error ||= "Warning: there is no color named \"#{ufg}\", using fallback."
-            warn "there is no color named \"#{ufg}\""
-          end
-        end
-
-        if(ubg = ucolor[:bg])
-          begin
-            bg = Curses.const_get "COLOR_#{ubg.upcase}"
-          rescue NameError
-            error ||= "Warning: there is no color named \"#{ubg}\", using fallback."
-            warn "there is no color named \"#{ubg}\""
-          end
-        end
-
-        if(uattrs = ucolor[:attrs])
-          attrs = [*uattrs].flatten.map do |a|
-            begin
-              Curses.const_get "A_#{a.upcase}"
-            rescue NameError
-              error ||= "Warning: there is no attribute named \"#{a}\", using fallback."
-              warn "there is no attribute named \"#{a}\", using fallback."
-            end
-          end
-        end
+    Colormap::DEFAULT_COLORS.merge(user_colors||{}).each_pair do |k, v|
+      fg = begin
+        Curses.const_get "COLOR_#{v[:fg].to_s.upcase}"
+      rescue NameError
+        warn "there is no color named \"#{v[:fg]}\""
+        Curses::COLOR_GREEN
       end
 
-      symbol = (k.to_s + "_color").to_sym
-      add symbol, fg, bg, attrs
-    end
+      bg = begin
+        Curses.const_get "COLOR_#{v[:bg].to_s.upcase}"
+      rescue NameError
+        warn "there is no color named \"#{v[:bg]}\""
+        Curses::COLOR_RED
+      end
 
-    BufferManager.flash error if error
+      attrs = (v[:attrs]||[]).map do |a|
+        begin
+          Curses.const_get "A_#{a.upcase}"
+        rescue NameError
+          warn "there is no attribute named \"#{a}\", using fallback."
+          nil
+        end
+      end.compact
+
+      highlight_symbol = v[:highlight] ? :"#{v[:highlight]}_color" : nil
+
+      symbol = (k.to_s + "_color").to_sym
+      add symbol, fg, bg, attrs, highlight_symbol
+    end
   end
 
   def self.instance; @@instance; end
