@@ -3,7 +3,7 @@ module Redwood
 class ThreadViewMode < LineCursorMode
   ## this holds all info we need to lay out a message
   class MessageLayout
-    attr_accessor :top, :bot, :prev, :next, :depth, :width, :state, :color, :star_color, :orig_new
+    attr_accessor :top, :bot, :prev, :next, :depth, :width, :state, :color, :star_color, :orig_new, :toggled_state
   end
 
   class ChunkLayout
@@ -35,6 +35,14 @@ Return value:
   which will be appended by sup.
 EOS
 
+  HookManager.register "publish", <<EOS
+Executed when a message or a chunk is requested to be published.
+Variables:
+     chunk: Redwood::Message or Redwood::Chunk::* to be published.
+Return value:
+  None.
+EOS
+
   register_keymap do |k|
     k.add :toggle_detailed_header, "Toggle detailed header", 'h'
     k.add :show_header, "Show full message header", 'H'
@@ -46,7 +54,9 @@ EOS
     k.add :edit_labels, "Edit or add labels for a thread", 'l'
     k.add :expand_all_quotes, "Expand/collapse all quotes in a message", 'o'
     k.add :jump_to_next_open, "Jump to next open message", 'n'
+    k.add :jump_to_next_and_open, "Jump to next message and open", "\C-n"
     k.add :jump_to_prev_open, "Jump to previous open message", 'p'
+    k.add :jump_to_prev_and_open, "Jump to previous message and open", "\C-p"
     k.add :align_current_message, "Align current message in buffer", 'z'
     k.add :toggle_starred, "Star or unstar message", '*'
     k.add :toggle_new, "Toggle unread/read status of message", 'N'
@@ -59,6 +69,7 @@ EOS
     k.add :edit_as_new, "Edit message as new", 'D'
     k.add :save_to_disk, "Save message/attachment to disk", 's'
     k.add :save_all_to_disk, "Save all attachments to disk", 'A'
+    k.add :publish, "Publish message/attachment using publish-hook", 'P'
     k.add :search, "Search for messages from particular people", 'S'
     k.add :compose, "Compose message to person", 'm'
     k.add :subscribe_to_list, "Subscribe to/unsubscribe from mailing list", "("
@@ -74,6 +85,7 @@ EOS
       kk.add :delete_and_kill, "Delete this thread and kill buffer", 'd'
       kk.add :spam_and_kill, "Mark this thread as spam and kill buffer", 's'
       kk.add :unread_and_kill, "Mark this thread as unread and kill buffer", 'N'
+      kk.add :do_nothing_and_kill, "Just kill this buffer", '.'
     end
 
     k.add_multi "(a)rchive/(d)elete/mark as (s)pam/mark as u(N)read/do (n)othing:", ',' do |kk|
@@ -81,7 +93,7 @@ EOS
       kk.add :delete_and_next, "Delete this thread, kill buffer, and view next", 'd'
       kk.add :spam_and_next, "Mark this thread as spam, kill buffer, and view next", 's'
       kk.add :unread_and_next, "Mark this thread as unread, kill buffer, and view next", 'N'
-      kk.add :do_nothing_and_next, "Kill buffer, and view next", 'n'
+      kk.add :do_nothing_and_next, "Kill buffer, and view next", 'n', ','
     end
 
     k.add_multi "(a)rchive/(d)elete/mark as (s)pam/mark as u(N)read/do (n)othing:", ']' do |kk|
@@ -89,7 +101,7 @@ EOS
       kk.add :delete_and_prev, "Delete this thread, kill buffer, and view previous", 'd'
       kk.add :spam_and_prev, "Mark this thread as spam, kill buffer, and view previous", 's'
       kk.add :unread_and_prev, "Mark this thread as unread, kill buffer, and view previous", 'N'
-      kk.add :do_nothing_and_prev, "Kill buffer, and view previous", 'n'
+      kk.add :do_nothing_and_prev, "Kill buffer, and view previous", 'n', ']'
     end
   end
 
@@ -101,7 +113,7 @@ EOS
   ## objects. @person_lines is a map from row #s to Person objects.
 
   def initialize thread, hidden_labels=[], index_mode=nil
-    super()
+    super :slip_rows => $config[:slip_rows]
     @thread = thread
     @hidden_labels = hidden_labels
 
@@ -119,6 +131,7 @@ EOS
       next unless m
       earliest ||= m
       @layout[m].state = initial_state_for m
+      @layout[m].toggled_state = false
       @layout[m].color = altcolor ? :alternate_patina_color : :message_patina_color
       @layout[m].star_color = altcolor ? :alternate_starred_patina_color : :starred_patina_color
       @layout[m].orig_new = m.has_label? :read
@@ -330,7 +343,7 @@ EOS
     elsif chunk.viewable?
       view chunk
     end
-    if chunk.is_a?(Message)
+    if chunk.is_a?(Message) && $config[:jump_to_open_message]
       jump_to_message chunk
       jump_to_next_open if layout.state == :closed
     end
@@ -388,6 +401,15 @@ EOS
     end
   end
 
+  def publish
+    chunk = @chunk_lines[curpos] or return
+    if HookManager.enabled? "publish"
+      HookManager.run "publish", :chunk => chunk
+    else
+      BufferManager.flash "Publishing hook not defined."
+    end
+  end
+
   def edit_draft
     m = @message_lines[curpos] or return
     if m.is_draft?
@@ -421,6 +443,29 @@ EOS
     end
   end
 
+  def jump_to_next_and_open
+    return continue_search_in_buffer if in_search? # err.. don't know why im doing this
+
+    m = (curpos ... @message_lines.length).argfind { |i| @message_lines[i] }
+    return unless m
+
+    if @layout[m].toggled_state == true
+      @layout[m].state = :closed
+      @layout[m].toggled_state = false
+      update
+    end
+
+    nextm = @layout[m].next
+    if @layout[nextm].state == :closed
+      @layout[nextm].state = :open
+      @layout[nextm].toggled_state = true
+    end
+
+    jump_to_message nextm if nextm
+
+    update if @layout[nextm].toggled_state
+  end
+
   def jump_to_next_open force_alignment=nil
     return continue_search_in_buffer if in_search? # hack: allow 'n' to apply to both operations
     m = (curpos ... @message_lines.length).argfind { |i| @message_lines[i] }
@@ -437,6 +482,26 @@ EOS
     jump_to_message m, true
   end
 
+  def jump_to_prev_and_open force_alignment=nil
+    m = (0 .. curpos).to_a.reverse.argfind { |i| @message_lines[i] }
+    return unless m
+
+    if @layout[m].toggled_state == true
+      @layout[m].state = :closed
+      @layout[m].toggled_state = false
+      update
+    end
+
+    nextm = @layout[m].prev
+    if @layout[nextm].state == :closed
+      @layout[nextm].state = :open
+      @layout[nextm].toggled_state = true
+    end
+
+    jump_to_message nextm if nextm
+    update if @layout[nextm].toggled_state
+  end
+
   def jump_to_prev_open
     m = (0 .. curpos).to_a.reverse.argfind { |i| @message_lines[i] } # bah, .to_a
     return unless m
@@ -508,6 +573,7 @@ EOS
   def spam_and_kill; spam_and_then :kill end
   def delete_and_kill; delete_and_then :kill end
   def unread_and_kill; unread_and_then :kill end
+  def do_nothing_and_kill; do_nothing_and_then :kill end
 
   def archive_and_next; archive_and_then :next end
   def spam_and_next; spam_and_then :next end
@@ -792,7 +858,12 @@ private
       if chunk.inlineable?
         lines = chunk.lines
         if @wrap
-          width = buffer.content_width
+          config_width = $config[:wrap_width]
+          if config_width and config_width != 0
+            width = [config_width, buffer.content_width].min
+          else
+            width = buffer.content_width
+          end
           lines = lines.map { |l| l.chomp.wrap width }.flatten
         end
         lines.map { |line| [[chunk.color, "#{prefix}#{line}"]] }
