@@ -85,6 +85,7 @@ class Thread
   end
 
   def first; each { |m, *o| return m if m }; nil; end
+  def has_message?; any? { |m, *o| m.is_a? Message }; end
   def dirty?; any? { |m, *o| m && m.dirty? }; end
   def date; map { |m, *o| m.date if m }.compact.max; end
   def snippet
@@ -146,6 +147,11 @@ class Thread
   def to_s
     "<thread containing: #{@containers.join ', '}>"
   end
+
+  def sort_key
+    m = latest_message
+    m ? [-m.date.to_i, m.id] : [-Time.now.to_i, ""]
+  end
 end
 
 ## recursive structure used internally to represent message trees as
